package com.supervision.pdfqaserver.dao;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.supervision.pdfqaserver.dto.EntityExtractionDTO;
import com.supervision.pdfqaserver.dto.RelationExtractionDTO;
import com.supervision.pdfqaserver.dto.TruncationERAttributeDTO;
import com.supervision.pdfqaserver.dto.neo4j.NodeData;
import com.supervision.pdfqaserver.dto.neo4j.RelationObject;
import com.supervision.pdfqaserver.dto.neo4j.RelationshipData;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.neo4j.driver.*;
import org.neo4j.driver.Record;
import org.neo4j.driver.types.Node;
import org.neo4j.driver.types.Relationship;
import org.springframework.stereotype.Repository;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;
import static org.neo4j.driver.Values.parameters;

@Slf4j
@Repository
@RequiredArgsConstructor
public class Neo4jRepository {
    /**
     * Neo4j 驱动
     */
    private final Driver driver;

    /**
     * 执行传入的 Cypher 语句并返回 RelationObject 列表
     *
     * @param cypher  完整的预生成 Cypher 语句，比如：
     *                "MATCH (a)-[r:REL_TYPE]->(b) RETURN a AS startNode, r AS rel, b AS endNode"
     * @param params  语句参数（可为空）
     */
    public List<RelationObject> execute(String cypher, Map<String, Object> params) {
        try (Session session = driver.session()) {
            return session.executeRead(tx -> {
                Result result = tx.run(cypher, params == null ? Collections.emptyMap() : params);
                List<RelationObject> list = new ArrayList<>();
                while (result.hasNext()) {
                    org.neo4j.driver.Record record = result.next();
                    // 从 Record 中取出三部分
                    if (record.get("c") == null || record.get("r") == null || record.get("t") == null) {
                        continue;
                    }
                    Node a = record.get("c").asNode();
                    Relationship r = record.get("r").asRelationship();
                    Node b = record.get("t").asNode();

                    // 转成我们的 DTO
                    NodeData start = mapNode(a);
                    RelationshipData rel = mapRel(r);
                    NodeData end = mapNode(b);

                    list.add(new RelationObject(start, rel, end));
                }
                return list;
            });
        }
    }


    /**
     * 执行原生 Cypher 语句
     * @param cypher 原生 Cypher 语句
     * @param params 参数
     * @return List<Record>
     */
    public List<Record> executeCypherNative(String cypher, Map<String, Object> params) {
        try (Session session = driver.session()) {
            Result run = session.run(cypher, params == null ? Collections.emptyMap() : params);
            return run.list();
        }
    }


    /**
     * 创建或更新实体节点
     * 根据唯一键（uniqueKey）来判断节点是否存在
     * note: properties中需要包含唯一键的值
     * @param label label
     * @param uniqueKey 唯一键
     * @param properties 属性
     * @return 节点ID列表
     */
    public List<Long> saveOrUpdateEntityNode(String label,String uniqueKey,Map<String, Object> properties) {

        try (Session session = driver.session()) {
            // MERGE语句确保唯一性
            String query = String.format(
                    "MERGE (n:%s {%s: $uniqueValue}) " +
                            "SET n += $properties " +
                            "RETURN id(n)  as id ",
                    label, uniqueKey);

            Result result = session.run(query, parameters("uniqueValue", properties.get(uniqueKey), "properties", properties));
            return result.list().stream().map(record -> record.get("id").asLong()).collect(Collectors.toList());
        }
    }

    /**
     * 创建或更新关系
     * @param sourceId 头节点ID
     * @param targetId 尾节点ID
     * @param relationType 关系类型
     * @param leftDirection 方向左
     * @param rightDirection 方向右
     * @param properties 关系属性
     * @return
     */
    public List<Long> saveOrUpdateRelation(Long sourceId,Long targetId,String relationType,
                                            boolean leftDirection,boolean rightDirection,Map<String, Object> properties) {

        try (Session session = driver.session()) {
            // MERGE关系
            String query = "MATCH (a) WHERE id(a) = $sourceId " +
                    "MATCH (b) WHERE id(b) = $targetId " +
                    "MERGE (a)%s-[r:" + relationType + "]-%s(b) " +
                    "SET r += $properties " +
                    "RETURN id(r) as id";

            String format = String.format(query,
                    leftDirection ? "<" : "",
                    rightDirection ? ">" : "");

            Result result = session.run(format,
                    parameters("sourceId", sourceId, "targetId", targetId, "properties", properties));

            return result.stream().map(record -> record.get("id").asLong()).collect(Collectors.toList());
        }
    }

    /**
     * 获取节点的schema
     * @return
     */
    public List<EntityExtractionDTO> getNodeSchema(){

        String query = """
                CALL db.schema.nodeTypeProperties()
                    YIELD nodeType, propertyName, propertyTypes
                    RETURN nodeType, propertyName, propertyTypes
                """;
        try (Session session = driver.session()) {

            List<EntityExtractionDTO> extractionDTOS = new ArrayList<>();
            Result result = session.run(query);
            for (Record record : result.list()) {
                String nodeType = record.get("nodeType").asString();
                if (StrUtil.isEmpty(nodeType)){
                    continue;
                }
                nodeType = nodeType.substring(1, nodeType.length()-1).replace("`", "");
                String propertyName = record.get("propertyName").asString();
                List<String> propertyTypes = record.get("propertyTypes").asList(Value::asString);

                // 创建属性DTO
                TruncationERAttributeDTO attributeDTO = new TruncationERAttributeDTO(propertyName, null, CollUtil.getFirst(propertyTypes));

                // 检查是否已存在该节点类型
                final String nodeType_f = nodeType;
                EntityExtractionDTO existingEntity = extractionDTOS.stream()
                        .filter(e -> StrUtil.equals(e.getEntityEn(), nodeType_f))
                        .findFirst().orElse(null);

                if (existingEntity != null) {
                    // 如果已存在，添加属性
                    existingEntity.getAttributes().add(attributeDTO);
                } else {
                    // 如果不存在，创建新的实体DTO
                    List<TruncationERAttributeDTO> truncationERAttributeDTOS = new ArrayList<>();
                    truncationERAttributeDTOS.add(attributeDTO);
                    EntityExtractionDTO entityExtractionDTO = new EntityExtractionDTO(null,nodeType,  null,truncationERAttributeDTOS);
                    extractionDTOS.add(entityExtractionDTO);
                }
            }
            return extractionDTOS;
        }
    }

    /**
     * 获取关系的schema
     * @return
     */
    public List<RelationExtractionDTO> getRelationSchema(){
        String queryProper = """
                 CALL db.schema.relTypeProperties()
                    YIELD relType, propertyName, propertyTypes
                    RETURN relType, propertyName, propertyTypes
                """;
        Map<String, List<Map<String, String>>> relationProperties = new HashMap<>();
        try (Session session = driver.session()) {
            Result result = session.run(queryProper);
            for (Record record : result.list()) {
                String relType = record.get("relType").asString();
                if (StrUtil.isEmpty(relType)){
                    continue;
                }
                relType = relType.substring(1, relType.length()-1).replace("`", "");
                String propertyName = record.get("propertyName").asString();
                List<String> propertyTypes = record.get("propertyTypes").asList(Value::asString);

                List<Map<String, String>> properties = relationProperties.computeIfAbsent(relType, k -> new ArrayList<>());
                boolean noneMatch = properties.stream().noneMatch(
                        prop -> StrUtil.equals(prop.get("propertyName"), propertyName)
                );
                if (noneMatch){
                    Map<String, String> propMap = new HashMap<>();
                    propMap.put("propertyName", propertyName);
                    propMap.put("propertyTypes", CollUtil.getFirst(propertyTypes));
                    properties.add(propMap);
                }
            }

            List<RelationExtractionDTO> relationExtractionDTOS = new ArrayList<>();
            String queryEndpoints = """
                MATCH (s)-[r: `{rtype}` ]->(t)
                        WITH labels(s)[0] AS src, labels(t)[0] AS tgt
                        RETURN src, tgt
                """;
            for (Map.Entry<String, List<Map<String, String>>> entry : relationProperties.entrySet()) {
                String relType = entry.getKey();
                List<Map<String, String>> properties = entry.getValue();
                String formatted = StrUtil.format(queryEndpoints,Map.of("rtype",relType));
                Result run = session.run(formatted);
                for (Record record : run.list()) {
                    String sourceType = record.get("src").asString();
                    String targetType = record.get("tgt").asString();
                    List<TruncationERAttributeDTO> attributeDTOS = properties.stream().map(
                            prop -> new TruncationERAttributeDTO(prop.get("propertyName"), null, prop.get("propertyTypes"))
                    ).collect(Collectors.toList());
                    RelationExtractionDTO relationExtractionDTO = new RelationExtractionDTO(null,null,sourceType,
                            relType,
                            null,
                            targetType,
                            attributeDTOS);
                    relationExtractionDTOS.add(relationExtractionDTO);
                }
            }
            return relationExtractionDTOS;
        }
    }


    private NodeData mapNode(Node node) {
        return new NodeData(
                node.id(),
                StreamSupport.stream(node.labels().spliterator(), false).collect(Collectors.toList()),
                node.asMap()
        );
    }

    private RelationshipData mapRel(Relationship r) {
        return new RelationshipData(
                r.id(),
                r.type(),
                r.asMap()
        );
    }
}
